// FastCode / ScaleDown function / High-Precision FPU version
//    Eric Grange

unit ScaleEG_FPU_Unit;

interface

uses Classes, QGraphics, Types;

function ScaleEG_FPUPas(const SourceBitmap : TBitmap; ScaleFactor : Single) :
TBitmap;
procedure LinearResample(srcBmp32 : TBitmap; destBmp32 : TBitmap);

var
   v3DNowAllowed : Boolean;

implementation

type
   //
   // Small wrapper classes around TBitmap.ScanLine, 24 & 32bits pixel format support
   //

   TBMPPixelWrapper = class
      FScanLines : array of Pointer;
      FBitmap : TBitmap;
      constructor Create(aBitmap : TBitmap); virtual;
      function GetPixel(x, y : Integer) : TColor; virtual; abstract;
      procedure SetPixel(x, y : Integer; col : TColor); virtual; abstract;
   end;


   TBMPPW32Bits = class (TBMPPixelWrapper)
      constructor Create(aBitmap : TBitmap); override;
      function GetPixel(x, y : Integer) : TColor; override;
      procedure SetPixel(x, y : Integer; col : TColor); override;
   end;

constructor TBMPPixelWrapper.Create(aBitmap : TBitmap);
var
   i : Integer;
   startScanLine, delta : Integer;
begin
   FBitmap:=aBitmap;
   SetLength(FScanLines, FBitmap.Height);
   case FBitmap.Height of
      0 : ;
      1 : FScanLines[0]:=FBitmap.ScanLine[0];
   else
      startScanLine:=Integer(FBitmap.ScanLine[0]);
      delta:=Integer(FBitmap.ScanLine[1])-startScanLine;
      for i:=0 to FBitmap.Height-1 do
         FScanLines[i]:=Pointer(startScanLine+i*delta);
   end;
end;







constructor TBMPPW32Bits.Create(aBitmap : TBitmap);
begin
   Assert(aBitmap.PixelFormat=pf32bit);
   inherited;
end;

function TBMPPW32Bits.GetPixel(x, y : Integer) : TColor;
begin
     {Assert((Cardinal(x)<Cardinal(FBitmap.Width)) and
     (Cardinal(y)<Cardinal(FBitmap.Height)), 'Incorrect x, y : '+IntToStr(x)+',
     '+IntToStr(y));}
   Result:=PInteger(@PChar(FScanLines[y])[x*4])^;
end;

procedure TBMPPW32Bits.SetPixel(x, y : Integer; col : TColor);
begin
     {//   Assert((Cardinal(x)<Cardinal(FBitmap.Width)) and
         (Cardinal(y)<Cardinal(FBitmap.Height)), 'Incorrect x, y : '+IntToStr(x)+',
          '+IntToStr(y)); }
   PInteger(@PChar(FScanLines[y])[x*4])^:=col;
end;

type
   TLeftRightWeights = packed record
      left, leftDummy : Single;
      right, rightDummy : Single;
   end;

// 3DNow! Accelerated source block accumulation
//        (about 3x faster on AXP)
function BlockAccumulate(bmpW : TBMPPixelWrapper;
                         const rect : TRect;
                         topWeight, bottomWeight : Single;
                         var lrWeight : TLeftRightWeights;
                         invSum : Single) : TColor;
type
   TIntegerArray = array [0..MaxInt shr 3] of Integer;
   PIntegerArray = ^TIntegerArray;
   PColor = ^TColor;

   procedure Accumulate(color : PColor);
   // expects RGBA accumulators in MM0:MM1
   //         weight in MM6
   //         zero in MM7
   asm
      prefetch    [edx+24]
      movd        mm2, [edx]     // load rgba components into mm2:mm3 as 32bits integers
      movd        mm3, [edx]
      punpcklbw   mm2, mm7
      punpcklbw   mm3, mm7
      punpcklwd   mm2, mm7
      punpckhwd   mm3, mm7
      pi2fd       mm2, mm2       // convert rgba components to floating point
      pi2fd       mm3, mm3
      pfmul       mm2, mm6       // scale by weight
      pfmul       mm3, mm6
      pfadd       mm0, mm2       // accumulate
      pfadd       mm1, mm3
   end;

   procedure AccumulateRow(scanLine : PIntegerArray;
                           const rect : TRect;
                           lrWeight : TLeftRightWeights);
   // expects RGBA accumulators in MM0:MM1
   //         weight in MM5
   //         zero in MM7
   asm
      push  esi
      push  ecx

      mov   esi, edx

      movq  mm6, [ecx] // lrWeight.left
      pfmul mm6, mm5

      mov   ecx, [esi] // rect.left
      lea   edx, [eax+ecx*4]
      call  Accumulate

      mov   eax, [esi+8] // rect.right
      sub   eax, ecx

      movq  mm6, mm5

   @@Loop:
      lea   edx, [edx+4]
      dec   eax
      jz   @@LoopEnd

      call  Accumulate
      jmp   @@Loop

   @@LoopEnd:

      pop   ecx

      movq  mm6, [ecx+8]  // lrWeight.right
      pfmul mm6, mm5

      call  Accumulate

      pop   esi
   end;

const
   cOneOne : array [0..1] of Single = (1, 1);
var
   y : Integer;
   pp : PPointerArray;
begin

   asm
      pxor  mm0, mm0    // RAZ RGBA accumulator
      pxor  mm1, mm1

      pxor  mm7, mm7    // global zero
   end;

   // top line
   asm
      movd        mm5, topWeight
      punpckldq   mm5, mm5
   end;
   AccumulateRow(pp[rect.Top], rect, lrWeight);
   // mi
   asm
      movq        mm5, cOneOne
   end;
   for y:=rect.Top+1 to rect.Bottom-1 do
      AccumulateRow(pp[y], rect, lrWeight);
   // bottom line
   asm
      movd        mm5, bottomWeight
      punpckldq   mm5, mm5
   end;
   AccumulateRow(pp[rect.Bottom], rect, lrWeight);

   // compute resulting color
   asm
      movd        mm6, invSum    // load invSum
      punpckldq   mm6, mm6

      pfmul       mm0, mm6       // scale rgba components
      pfmul       mm1, mm6

      pf2id       mm0, mm0       // back to 32bits integers
      pf2id       mm1, mm1

      packssdw    mm0, mm1       // pack into 8bits integers
      packuswb    mm0, mm0
      movd        y, mm0

      femms
   end;
   Result:=y;
end;

// Isotropic linear resampling, implementation notes:
// - supports 24 and 32 bits bitmaps, src and dest are allowed to differ
// - floating point precision
// - local filtering only for upsampling (low quality)
// - all 4 RGBA channels are resampled
procedure LinearResample(srcBmp32 : TBitmap; destBmp32 : TBitmap);
var
   rgba : array [0..3] of Single;
   srcBmpWrapper, destBmpWrapper : TBMPPixelWrapper;

   procedure Accumulate(sx, sy : Integer; const w : Single);
   var
      color : TColor;
   begin
      color:=srcBmpWrapper.GetPixel(sx, sy);// srcBmp32.Canvas.Pixels[sx, sy];

      rgba[0]:=rgba[0]+(color and $FF)*w;
      rgba[1]:=rgba[1]+((color shr 8) and $FF)*w;
      rgba[2]:=rgba[2]+((color shr 16) and $FF)*w;
      rgba[3]:=rgba[3]+((color shr 24) and $FF)*w;
   end;

   procedure AccumulateOne(color : TColor);
   begin
      rgba[0]:=rgba[0]+(color and $FF);
      rgba[1]:=rgba[1]+((color shr 8) and $FF);
      rgba[2]:=rgba[2]+((color shr 16) and $FF);
      rgba[3]:=rgba[3]+((color shr 24) and $FF);
   end;

   procedure BackupAndSetupCW(var cw : Integer);
   asm
      fnstcw  [eax]
      fnstcw  [eax+2]
      or      [eax+2], $0F00 // trunc mode
      fldcw   [eax+2]
   end;

   procedure RestoreCW(var cw : Integer);
   asm
      fldcw   [eax]
   end;

   function MyFrac(var s : Single) : Single;
   asm
      fld      dword ptr [eax]
      fld      dword ptr [eax]
      frndint
      fsub
   end;

   procedure TruncFrac(var s : Single; var truncRes : Integer; var fracRes :
Single);
   asm
      fld      dword ptr [eax]
      fist     dword ptr [edx]
      fld      dword ptr [eax]
      frndint
      fsub
      fstp     dword ptr [ecx]
   end;

const
   cAlmostOne : Single = 0.9999;
var
   x, y, sx, sy : Integer;
   xScale, yScale, weight, invSum : Single;
   srcXMin, srcXMax, srcYMin, srcYMax : Single;
   srcRect : TRect;
   fSrcXMin, fSrcXMax, fSrcYMin, fSrcYMax : Single;
   lrWeights : TLeftRightWeights;
   savedControlWord : Integer;
   canUse3DNow : Boolean;
begin
   // For each pixel in the destination bitmap:
   // - determine its coverage in the source bitmap
   // - accumulate weighted average of the source pixels
   // - issue pixel

   case srcBmp32.PixelFormat of
      pf32bit : srcBmpWrapper:=TBMPPW32Bits.Create(srcBmp32);
   else
      srcBmpWrapper:=nil;
      Assert(False);
   end;
   case destBmp32.PixelFormat of
      pf32bit : destBmpWrapper:=TBMPPW32Bits.Create(destBmp32);
   else
      destBmpWrapper:=nil;
      Assert(False);
   end;

   canUse3DNow:=False and v3DNowAllowed and (srcBmp32.PixelFormat=pf32bit);

   // we setup default rounding mode for the duration of the resampling
   BackupAndSetupCW(savedControlWord);

   // scale factors convert from destination coordinates to source coordinates
   xScale:=srcBmp32.Width/destBmp32.Width;
   yScale:=srcBmp32.Height/destBmp32.Height;
   invSum:=1/(xScale*yScale);

   for y:=0 to destBmp32.Height-1 do begin
      srcYMin:=y*yScale;
      srcYMax:=(y+cAlmostOne)*yScale;

      TruncFrac(srcYMin, srcRect.Top, fSrcYMin);
      TruncFrac(srcYMax, srcRect.Bottom, fSrcYMax);

      for x:=0 to destBmp32.Width-1 do begin
         srcXMin:=x*xScale;
         srcXMax:=(x+cAlmostOne)*xScale;

         TruncFrac(srcXMin, srcRect.Left, fSrcXMin);
         TruncFrac(srcXMax, srcRect.Right, fSrcXMax);

         // reset accumulator
         rgba[0]:=0;
         rgba[1]:=0;
         rgba[2]:=0;
         rgba[3]:=0;

         // Accumulate contributions
         if srcRect.Top=srcRect.Bottom then begin
            // src Y range within a single row (magnification)
            if srcRect.Left=srcRect.Right then begin
               // src X range within a single column (magnification)
               // no filtering applied, only point sampling right now
               destBmpWrapper.SetPixel(x, y,
srcBmpWrapper.GetPixel(srcRect.Left, srcRect.Top));
               continue;
            end else begin
               // accumulate 1st and last column pixels separately
               weight:=srcYMax-srcYMin;
               Accumulate(srcRect.Left, srcRect.Top, (1-fSrcXMin)*weight);
               for sx:=srcRect.Left+1 to srcRect.Right-1 do
                  Accumulate(sx, srcRect.Top, weight);
               Accumulate(srcRect.Right, srcRect.Top, fSrcXMax*weight);
            end;
         end else begin
            // source area overlaps at least two rows
            if srcRect.Left=srcRect.Right then begin
               // accumulate 1st and last row pixels separately
               weight:=srcXMax-srcXMin;
               Accumulate(srcRect.Left, srcRect.Top, weight*(1-fSrcYMin));
               for sy:=srcRect.Top+1 to srcRect.Bottom-1 do
                  Accumulate(srcRect.Left, sy, weight);
               Accumulate(srcRect.Left, srcRect.Bottom, weight*fSrcYMax);
            end else begin
               // source area overlaps several rows and columns
               if canUse3DNow then begin
                  lrWeights.left:=1-fSrcXMin;
                  lrWeights.leftDummy:=lrWeights.left;
                  lrWeights.right:=fSrcXMax;
                  lrWeights.rightDummy:=fSrcXMax;
                  destBmpWrapper.SetPixel(x, y,
                     BlockAccumulate(srcBmpWrapper, srcRect,
                                     1-fSrcYMin, fSrcYMax, lrWeights,
                                     invSum)
                  );
                  continue;
               end else begin
                  // 1st row
                  weight:=1-fSrcYMin;
                  Accumulate(srcRect.Left, srcRect.Top, weight*(1-fSrcXMin));
                  for sx:=srcRect.Left+1 to srcRect.Right-1 do
                     Accumulate(sx, srcRect.Top, weight);
                  Accumulate(srcRect.Right, srcRect.Top, weight*fSrcXMax);

                  // middle rows
                  for sy:=srcRect.Top+1 to srcRect.Bottom-1 do begin
                     Accumulate(srcRect.Left, sy, 1-fSrcXMin);
                     for sx:=srcRect.Left+1 to srcRect.Right-1 do
                        AccumulateOne(srcBmpWrapper.GetPixel(sx, sy));
                     Accumulate(srcRect.Right, sy, fSrcXMax);
                  end;

                  // last row (weight = fSrcYMax)
                  Accumulate(srcRect.Left, srcRect.Bottom,
fSrcYMax*(1-fSrcXMin));
                  for sx:=srcRect.Left+1 to srcRect.Right-1 do
                     Accumulate(sx, srcRect.Bottom, fSrcYMax);
                  Accumulate(srcRect.Right, srcRect.Bottom, fSrcYMax*fSrcXMax);
               end;
            end;
         end;

         // Issue pixel
         destBmpWrapper.SetPixel(x, y,  (Round(rgba[3]*invSum) shl 24)
                                       +(Round(rgba[2]*invSum) shl 16)
                                       +(Round(rgba[1]*invSum) shl 8)
                                       + Round(rgba[0]*invSum));
      end;
   end;

   // Restore original control word
   RestoreCW(savedControlWord);

   destBmpWrapper.Free;
   srcBmpWrapper.Free;
end;

// ScaleEG_FPUPas
//   (wrapper for FastCode call convention)
function ScaleEG_FPUPas(const SourceBitmap : TBitmap; ScaleFactor : Single) :
TBitmap;
var
   dstRowNoMax, dstColNoMax : Integer;
begin
   Assert(ScaleFactor>0);

   dstRowNoMax := Round(SourceBitmap.Height/ScaleFactor);
   if dstRowNoMax<1 then
      dstRowNoMax:=1;

   dstColNoMax := Round(SourceBitmap.Width/ScaleFactor);
   if dstColNoMax<1 then
      dstColNoMax:=1;

   Result:=TBitmap.Create;
   Result.PixelFormat:=SourceBitmap.PixelFormat;
   Result.Height := dstRowNoMax;
   Result.Width := DstColNoMax;

   LinearResample(SourceBitmap, Result);
end;

//--------------------------------------------------------------
//--------------------------------------------------------------
//--------------------------------------------------------------
initialization
//--------------------------------------------------------------
//--------------------------------------------------------------
//--------------------------------------------------------------

   v3DNowAllowed:=False;
   try
      // detect 3DNow! capable CPU (adapted from AMD's "3DNow! Porting Guide")
      asm
         pusha
         mov  eax, $80000000
         db $0F,$A2               /// cpuid
         cmp  eax, $80000000
         jbe @@No3DNow
         mov  eax, $80000001
         db $0F,$A2               /// cpuid
         test edx, $80000000
         jz @@No3DNow
         mov v3DNowAllowed, 1
@@No3DNow:
         popa
      end;
   except
      // trap for old/exotics CPUs
      v3DNowAllowed:=False;
   end;

end.
